@@ -35,14 +35,10 @@ try:
     from packaging.specifiers import SpecifierSet
     from packaging.version import Version
 except ImportError:
-    SpecifierSet = None
-    Version = None
+    SpecifierSet = None  # type: ignore[misc]
+    Version = None  # type: ignore[misc]
 
-try:
-    import importlib.util
-    imp = None
-except ImportError:
-    import imp
+import importlib.util
 
 display = Display()
 
@@ -129,6 +125,22 @@ class PluginLoadContext(object):
         self.removal_version = None
         self.deprecation_warnings = []
         self.resolved = False
+        self._resolved_fqcn = None
+
+    @property
+    def resolved_fqcn(self):
+        if not self.resolved:
+            return
+
+        if not self._resolved_fqcn:
+            final_plugin = self.redirect_list[-1]
+            if AnsibleCollectionRef.is_valid_fqcr(final_plugin) and final_plugin.startswith('ansible.legacy.'):
+                final_plugin = final_plugin.split('ansible.legacy.')[-1]
+            if self.plugin_resolved_collection and not AnsibleCollectionRef.is_valid_fqcr(final_plugin):
+                final_plugin = self.plugin_resolved_collection + '.' + final_plugin
+            self._resolved_fqcn = final_plugin
+
+        return self._resolved_fqcn
 
     def record_deprecation(self, name, deprecation, collection_name):
         if not deprecation:
@@ -433,7 +445,7 @@ class PluginLoader:
             entry = collection_meta.get('plugin_routing', {}).get(plugin_type, {}).get(subdir_qualified_resource, None)
         return entry
 
-    def _find_fq_plugin(self, fq_name, extension, plugin_load_context):
+    def _find_fq_plugin(self, fq_name, extension, plugin_load_context, ignore_deprecated=False):
         """Search builtin paths to find a plugin. No external paths are searched,
         meaning plugins inside roles inside collections will be ignored.
         """
@@ -452,7 +464,8 @@ class PluginLoader:
             deprecation = routing_metadata.get('deprecation', None)
 
             # this will no-op if there's no deprecation metadata for this plugin
-            plugin_load_context.record_deprecation(fq_name, deprecation, acr.collection)
+            if not ignore_deprecated:
+                plugin_load_context.record_deprecation(fq_name, deprecation, acr.collection)
 
             tombstone = routing_metadata.get('tombstone', None)
 
@@ -606,7 +619,8 @@ class PluginLoader:
                                                                        plugin_load_context, ignore_deprecated, check_aliases, suffix)
                     else:
                         # 'ansible.builtin' should be handled here. This means only internal, or builtin, paths are searched.
-                        plugin_load_context = self._find_fq_plugin(candidate_name, suffix, plugin_load_context=plugin_load_context)
+                        plugin_load_context = self._find_fq_plugin(candidate_name, suffix, plugin_load_context=plugin_load_context,
+                                                                   ignore_deprecated=ignore_deprecated)
 
                         # Pending redirects are added to the redirect_list at the beginning of _resolve_plugin_step.
                         # Once redirects are resolved, ensure the final FQCN is added here.
@@ -738,7 +752,8 @@ class PluginLoader:
         # last ditch, if it's something that can be redirected, look for a builtin redirect before giving up
         candidate_fqcr = 'ansible.builtin.{0}'.format(name)
         if '.' not in name and AnsibleCollectionRef.is_valid_fqcr(candidate_fqcr):
-            return self._find_fq_plugin(fq_name=candidate_fqcr, extension=suffix, plugin_load_context=plugin_load_context)
+            return self._find_fq_plugin(fq_name=candidate_fqcr, extension=suffix, plugin_load_context=plugin_load_context,
+                                        ignore_deprecated=ignore_deprecated)
 
         return plugin_load_context.nope('{0} is not eligible for last-chance resolution'.format(name))
 
@@ -769,15 +784,10 @@ class PluginLoader:
 
         with warnings.catch_warnings():
             warnings.simplefilter("ignore", RuntimeWarning)
-            if imp is None:
-                spec = importlib.util.spec_from_file_location(to_native(full_name), to_native(path))
-                module = importlib.util.module_from_spec(spec)
-                spec.loader.exec_module(module)
-                sys.modules[full_name] = module
-            else:
-                with open(to_bytes(path), 'rb') as module_file:
-                    # to_native is used here because imp.load_source's path is for tracebacks and python's traceback formatting uses native strings
-                    module = imp.load_source(to_native(full_name), to_native(path), module_file)
+            spec = importlib.util.spec_from_file_location(to_native(full_name), to_native(path))
+            module = importlib.util.module_from_spec(spec)
+            spec.loader.exec_module(module)
+            sys.modules[full_name] = module
         return module
 
     def _update_object(self, obj, name, path, redirected_names=None):
@@ -914,10 +924,16 @@ class PluginLoader:
             basename = os.path.basename(name)
 
             if basename == '__init__' or basename in _PLUGIN_FILTERS[self.package]:
+                # either empty or ignored by the module blocklist
+                continue
+
+            if basename == 'base' and self.package == 'ansible.plugins.cache':
+                # cache has legacy 'base.py' file, which is wrapper for __init__.py
                 continue
 
             if dedupe and basename in loaded_modules:
                 continue
+
             loaded_modules.add(basename)
 
             if path_only:
@@ -1017,10 +1033,10 @@ class Jinja2Loader(PluginLoader):
         # Instead, calling code deduplicates jinja2 plugin names when loading each file.
         kwargs['_dedupe'] = False
 
-        # TODO: move this to initalization and extract/dedupe plugin names in loader and offset this from
+        # TODO: move this to initialization and extract/dedupe plugin names in loader and offset this from
         # caller. It would have to cache/refresh on add_directory to reevaluate plugin list and dedupe.
         # Another option is to always prepend 'ansible.legac'y and force the collection path to
-        # load/find plugins, just need to check compatiblity of that approach.
+        # load/find plugins, just need to check compatibility of that approach.
         # This would also enable get/find_plugin for these type of plugins.
 
         # We have to instantiate a list of all files so that we can reverse the list.
@@ -1125,7 +1141,8 @@ def _does_collection_support_ansible_version(requirement_string, ansible_version
 
 def _configure_collection_loader():
     if AnsibleCollectionConfig.collection_finder:
-        display.warning('AnsibleCollectionFinder has already been configured')
+        # this must be a Python warning so that it can be filtered out by the import sanity test
+        warnings.warn('AnsibleCollectionFinder has already been configured')
         return
 
     finder = _AnsibleCollectionFinder(C.config.get_config_value('COLLECTIONS_PATHS'), C.config.get_config_value('COLLECTIONS_SCAN_SYS_PATH'))
